package datastructures;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import exceptions.UnknownBeliefLabelException;

public class BeliefPermutationIndex {
    private Map<String, HashMap<Boolean, HashSet<BeliefPermutation>>> index = new HashMap<String, HashMap<Boolean, HashSet<BeliefPermutation>>>();

    /**
     * cycle through all individual beliefs (a, not b, not c, d, etc) and add this permutation to all
     * individual indices. i.e. there is an index node for "a", one for "not b", one for "not c", etc.
     * 
     * NOTE: this code does NOT un-register the permutation from the NEGATION of a, not b, not c, d... It
     * assumes that the calling code is sane, and will clear the index when necessary.
     * 
     * @param permutation
     *            the permutation to add to all applicable indices
     */
    public void register(BeliefPermutation permutation) {
        Map<String, Boolean> beliefs = permutation.getBeliefs();
        for (String label : beliefs.keySet()) {
            Boolean booleanValue = beliefs.get(label);

            // make sure an index entry exists for this label
            HashMap<Boolean, HashSet<BeliefPermutation>> labelIndex = index.get(label);
            if (labelIndex == null) {
                labelIndex = new HashMap<Boolean, HashSet<BeliefPermutation>>();
                index.put(label, labelIndex);
            }

            // make sure an index entry exists for this label/boolean
            // combination
            HashSet<BeliefPermutation> indexNode = labelIndex.get(booleanValue);
            if (indexNode == null) {
                indexNode = new HashSet<BeliefPermutation>();
                labelIndex.put(booleanValue, indexNode);
            }

            indexNode.add(permutation);
        }
    }

    /**
     * start over with a new index
     */
    public void clear() {
        index = new HashMap<String, HashMap<Boolean, HashSet<BeliefPermutation>>>();
    }

    public Set<BeliefPermutation> getBeliefPermutations(String label, Boolean booleanValue)
            throws UnknownBeliefLabelException {
        // make a copy for security
        HashMap<Boolean, HashSet<BeliefPermutation>> labelMap = index.get(label);
        if (labelMap == null) {
            throw new UnknownBeliefLabelException(label);
        }
        HashSet<BeliefPermutation> result = labelMap.get(booleanValue);
        return new HashSet<BeliefPermutation>(result); // return a copy so the index can't be modified
    }
}
