package datastructures;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;


import exceptions.IllegalRevisionException;
import exceptions.QueryFormattingException;
import exceptions.UnknownBeliefLabelException;

import queries.Query;
import revisionOperators.RevisionConstraint;

/**
 * Implements a Cumulative Measure as in Weydert 1984. Stores global structure in a skip list
 * (ConcurrentSkipListMap) with value -> node(s), and the local structure is a HashMap (name -> value)
 * 
 * @author luke
 * 
 */
public class CumulativeMeasure {
    private Set<BeliefPermutation> allPermutations = new HashSet<BeliefPermutation>();
    private final BeliefPermutationIndex index = new BeliefPermutationIndex();

    public CumulativeMeasure() {
        // start with an empty permutation, so that add() does something
        // meaningful when adding the first labeled belief.
        allPermutations.add(new BeliefPermutation());
    }

    public void setBeliefPermutations(Collection<BeliefPermutation> permutations)
            throws IllegalRevisionException {
        index.clear();

        // creating a new HashSet object assures which concrete implementation is used,
        // plus it guarantees correct hashing in the case of updated BeliefPermutations
        // which may have changed their hash and invalidated the hash table.
        allPermutations = new HashSet<BeliefPermutation>();
        for (BeliefPermutation permutation : permutations) {
            allPermutations.add(permutation);
            index.register(permutation);
        }
        validate();
    }

    private void validate() throws IllegalRevisionException {
        Value value = Value.sum(allPermutations);
        if (!value.equals(new Value(0, 1.0))) {
            throw new IllegalRevisionException();
        }
    }

    public void add(String label) {
        index.clear();
        HashSet<BeliefPermutation> newPermutations = new HashSet<BeliefPermutation>();
        Collection<BeliefPermutation> permutationsWithTrueBeliefAdded = addBeliefValueToPermutations(label,
                Boolean.TRUE, allPermutations);
        Collection<BeliefPermutation> permutationsWithFalseBeliefAdded = addBeliefValueToPermutations(label,
                Boolean.FALSE, allPermutations);
        newPermutations.addAll(permutationsWithTrueBeliefAdded);
        newPermutations.addAll(permutationsWithFalseBeliefAdded);

        // because we add 2 new possibilities (true + false),
        // each one gets 1/2 of the original belief value.
        for (BeliefPermutation beliefPermutation : newPermutations) {
            beliefPermutation.scaleLocalBelief(0.5);
            index.register(beliefPermutation);
        }

        allPermutations = newPermutations;
    }

    private Collection<BeliefPermutation> addBeliefValueToPermutations(String label, Boolean booleanValue,
            Set<BeliefPermutation> originalPermutations) {
        LinkedList<BeliefPermutation> result = new LinkedList<BeliefPermutation>();
        for (BeliefPermutation permutation : originalPermutations) {
            BeliefPermutation newPermutation = permutation.copy();
            newPermutation.addBelief(label, booleanValue);
            result.add(newPermutation);
        }
        return result;
    }

    public void revise(String queryString) throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        RevisionConstraint constraint = RevisionConstraint.fromQueryString(queryString);
        constraint.revise(this);
    }

    public void reviseSafe(String queryString) throws QueryFormattingException, UnknownBeliefLabelException {
        try {
            revise(queryString);
        } catch (IllegalRevisionException e) {
            // ignore only these - not formatting exceptions
        }
    }

    public Value getValue(Query query) throws UnknownBeliefLabelException {
        Set<BeliefPermutation> permutations = query.filter(allPermutations, index);
        Value totalValue = Value.sum(permutations);
        return totalValue;
    }

    public Set<BeliefPermutation> getBeliefPermutationsCopy() {
        HashSet<BeliefPermutation> result = new HashSet<BeliefPermutation>();
        for (BeliefPermutation permutation : allPermutations) {
            result.add(permutation.copy());
        }
        return result;
    }

    public Set<BeliefPermutation> getBeliefPermutations(String label, Boolean booleanValue)
            throws UnknownBeliefLabelException {
        return index.getBeliefPermutations(label, booleanValue);
    }

    public BeliefPermutationIndex getIndex() {
        return index; // no protection other than 'final' - be careful!
    }

    public Set<BeliefPermutation> query(Query query) throws UnknownBeliefLabelException {
        return query.filter(allPermutations, index);
    }

    public Set<BeliefPermutation> query(String queryString) throws UnknownBeliefLabelException,
            QueryFormattingException {
        return query(new Query(queryString));
    }

    /**
     * when doing value-changing functions, the permutations must be re-hashed, because the hash value is
     * computed based partially on the value, so key lookups and other functions will no longer work.
     */
    public void rehash() {
        allPermutations = new HashSet<BeliefPermutation>(allPermutations);
    }

    @Override
    public String toString() {
        return allPermutations.toString();
    }

}
