package datastructures;

import java.util.Collection;

import exceptions.QueryFormattingException;

public class Value {
    // initialize to (infinity, 0) to default to the "least likely" belief value
    public static final Integer GLOBAL_INFINITY = Integer.MAX_VALUE;
    public static final Double EQUALITY_THRESHOLD = 0.001; // workaround for floating-point errors and less verbose test cases

    Integer global = GLOBAL_INFINITY;
    Double local = 0.0;

    public static Value max() {
        return new Value(0, 1.0);
    }

    public static Value min() {
        return new Value(GLOBAL_INFINITY, 0.0);
    }

    public static Value sum(Collection<BeliefPermutation> permutations) {
        Value result = Value.min();
        for (BeliefPermutation permutation : permutations) {
            result.add(permutation.valueCopy());
        }
        return result;
    }

    public Value() {
    }

    public Value(Integer globalRankingValue, Double localRankingValue) {
        this.global = globalRankingValue;
        this.local = localRankingValue;
    }

    public Value(String valueString) throws QueryFormattingException {
        // examples of what might be passed in:
        // (1000,0.000)
        // (0,0)
        // (1,1.5)
        valueString = valueString.trim();
        if (valueString.charAt(0) != '(' || valueString.charAt(valueString.length() - 1) != ')') {
            throw new QueryFormattingException(valueString);
        }
        try {
            String globalString = valueString.split(",")[0].substring(1).trim();
            if (globalString.toLowerCase().equals("inf")) {
                this.global = GLOBAL_INFINITY;
            }
            else {
                this.global = new Integer(globalString);
            }
        }
        catch (Exception e) {
            throw new QueryFormattingException(valueString);
        }
        valueString = valueString.split(",")[1];
        valueString = valueString.substring(0, valueString.length() - 1); // remove the )
        this.local = new Double(valueString);
    }

    public boolean isValid() {
        if (local < 0.0 || global < 0 || (global == 0 && local > 1.0)) {
            return false;
        }
        return true;
    }

    public boolean greaterThan(Value other) {
        if (this.global.equals(other.global)) {
            // local value scale is normal (1.0 > 0.0)
            return (local > other.local + EQUALITY_THRESHOLD);
        }
        if (this.global < other.global) {
            // global ranking value scale is inverted (0 > INT_MAX)
            return true;
        }
        // this.globalRankingValue > value.globalRankingValue
        return false;
    }

    public boolean lessThan(Value other) {
        if (this.global.equals(other.global)) {
            // local value scale is normal (1.0 > 0.0)
            return (local < other.local - EQUALITY_THRESHOLD);
        }
        if (this.global > other.global) {
            // global ranking value scale is inverted (0 > INT_MAX)
            return true;
        }
        // this.globalRankingValue < value.globalRankingValue
        return false;
    }

    @Override
    public boolean equals(Object other) {
        if (!(other instanceof Value)) {
            return false;
        }
        Integer otherGlobal = ((Value) other).global;
        if (!otherGlobal.equals(global)) {
            return false;
        }
        Double otherLocal = ((Value) other).local;
        if (otherLocal > local + EQUALITY_THRESHOLD) {
            return false;
        }
        if (otherLocal < local - EQUALITY_THRESHOLD) {
            return false;
        }
        return true;
    }

    /**
     * do an addition operation with the other global & local values.
     * 
     * NOTE: assumes the caller is doing sanity checks (i.e. local values within rank 0 add up to 1)
     * 
     * @param other
     */
    public Value plus(Value other) {
        Value result = new Value(global, local);
        result.add(other);
        return result;
    }

    private void add(Value other) {
        if (this.global.equals(other.global)) {
            local += other.local;
        }
        else if (Math.abs(this.global) > Math.abs(other.global)) { // abs(): hack to work around INT.MAX being ambiguously signed
            // global ranking value scale is inverted (0 > INT_MAX)
            global = other.global;
            local = other.local;
        }
    }

    public Value dividedBy(Value other) {
        Value result = new Value(global, local);
        result.divideBy(other);
        return result;
    }

    private void divideBy(Value other) {
        global -= other.global;
        local /= other.local;
    }

    public Value multipliedBy(Value other) {
        Value result = new Value(global, local);
        result.multiplyBy(other);
        return result;
    }

    private void multiplyBy(Value other) {
        global += other.global;
        if (global > GLOBAL_INFINITY || global < 0) {
            global = GLOBAL_INFINITY;
        }
        local *= other.local;
    }

    public Integer global() {
        return global;
    }

    public void setGlobal(Integer newGlobal) {
        global = newGlobal;
    }

    public Double local() {
        return local;
    }

    public void scaleLocalBelief(Double scalingFactor) {
        local = local * scalingFactor;
    }

    @Override
    public String toString() {
        return String.format("(%d, %.3f)", global, local);
    }

    public Value copy() {
        return new Value(global, local);
    }
}
