package queries;

import exceptions.QueryFormattingException;
import exceptions.QueryParenthesesException;
import queries.QueryToken.EndGroup;
import queries.QueryToken.StartGroup;
import queries.QueryToken.UnknownToken;

public class QueryStringParser {
    /**
     * parse a query string and return the top node of a parse tree
     * 
     * @param criteriaString
     *            in the form "A AND b AND -c AND(D1 OR -d2 OR 3D) AND x"
     * @return parse tree representing grouped operations (parenthetical terms)
     */
    public static QueryToken parseCriteriaString(String criteriaString) throws QueryFormattingException {
        // split on spaces, left, and right parentheses, and create a LinkedList.
        TokenList list = new TokenList();
        criteriaString = criteriaString.trim().replaceAll("\\s+", " "); // collapse all whitespace down to single-space chars
        list.addAll(criteriaString.split(" "));
        explode(list, "(");
        explode(list, ")");

        // extract keywords & symbols
        extractKeywords(list, "AND", QueryToken.And.class);
        extractKeywords(list, "OR", QueryToken.Or.class);
        extractKeywords(list, "(", QueryToken.StartGroup.class);
        extractKeywords(list, ")", QueryToken.EndGroup.class);
        extractBeliefs(list);

        // create tree structure
        groupByParentheses(list);
        QueryToken topNode;
        try {
            topNode = list.asTree();
        }
        catch (QueryFormattingException e) {
            // add more description to the exception if needed.
            if (e.getMessage() == null) {
                QueryFormattingException newException = new QueryFormattingException(criteriaString);
                newException.addSuppressed(e);
                throw newException;
            }
            else {
                throw e;
            }
        }

        return topNode;
    }

    private static void explode(TokenList list, String string) {
        TokenListNode node = list.first;
        while (node != null) {
            TokenListNode splitTokens = node.split(string);
            try {
                list.insertNodes(node.token, splitTokens);
                list.delete(node.token);
            }
            catch (TokenNotFoundException e) {
                // shouldn't see me!
                e.printStackTrace();
            }
            node = node.next;
        }
    }

    private static void extractBeliefs(TokenList list) throws QueryFormattingException {
        TokenListNode node = list.first;
        while (node != null) {
            if (node.token instanceof UnknownToken) {
                QueryToken token = extractBeliefToken(((UnknownToken) node.token).toString());
                node.token = token;
            }
            node = node.next;
        }
    }

    public static QueryToken extractBeliefToken(String label) throws QueryFormattingException {
        label = label.trim();
        if (!label.matches("-?\\w+")) {
            throw new QueryFormattingException("label: " + label);
        }
        QueryToken token = new QueryToken.Belief(label);
        return token;
    }

    private static void extractKeywords(TokenList list, String keyword, Class<? extends QueryToken> tokenClass) {
        TokenListNode node = list.first;
        while (node != null) {
            if (node.token instanceof UnknownToken) {
                if (((UnknownToken) node.token).toString().equals(keyword)) {
                    // matched the token on the keyword passed in.
                    QueryToken newToken = null;
                    try {
                        newToken = tokenClass.newInstance();
                    }
                    catch (InstantiationException | IllegalAccessException e) {
                        // shouldn't see me!
                        e.printStackTrace();
                    }
                    node.token = newToken;
                }
            }
            node = node.next;
        }
    }

    private static void groupByParentheses(TokenList list) throws QueryFormattingException {
        groupByParentheses(list, false);
    }

    private static TokenListNode groupByParentheses(TokenList list, boolean shouldFindClosingParentheses)
            throws QueryFormattingException {
        TokenListNode node = list.first;
        while (node != null) {
            if (node.token instanceof StartGroup) {
                // we're at the next ( in the token list.

                if (node.next == null) {
                    // there was a ( at the end of a query
                    throw new QueryParenthesesException(list);
                }
                node.next.previous = null; // because it's now the start of a sub-list

                // replace the (...) sub-list with a GROUP node
                TokenListNodeGroup nodeGroup = new TokenListNodeGroup();
                nodeGroup.child = new TokenList();
                nodeGroup.child.add(node.next); // adds the entire remaining list (will be subdivided later)
                nodeGroup.previous = node.previous;
                nodeGroup.next = null;
                if (nodeGroup.previous == null) {
                    // we're at the start of the list, so we need to move the 'first' pointer forward to skip the (
                    list.first = nodeGroup;
                }
                else {
                    // we're not at the first node, so set the appropriate pointers
                    nodeGroup.previous.next = nodeGroup;
                }

                // recurse on the child level. Result of the call is the next sibling of this level.
                node = groupByParentheses(nodeGroup.child, true);

                if (node == null) {
                    // got to the end of the query string.
                    list.last = nodeGroup;
                    break;
                }

                nodeGroup.next = node;
                node.previous = nodeGroup;
            }
            else if (node.token instanceof EndGroup) {
                // closing a parentheses group.
                if (!shouldFindClosingParentheses) {
                    // too much ) and not enough (
                    throw new QueryParenthesesException(list);
                }
                // break the previous/next links and return the next node to process
                node.previous.next = null;
                if (node.next != null) {
                    node.next.previous = null;
                }
                list.last = node.previous;
                return node.next;
            }
            else {
                // nothing to see here.
                list.last = node;
                node = node.next;
            }
        }

        // got to the end of the list. 
        if (shouldFindClosingParentheses) {
            // too much ( and not enough )
            throw new QueryParenthesesException(list);
        }
        return null;
    }
}
