package queries;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import datastructures.BeliefPermutation;
import datastructures.BeliefPermutationIndex;
import exceptions.UnknownBeliefLabelException;

public abstract class QueryToken {
    protected Set<QueryToken> children = new HashSet<QueryToken>();

    public QueryToken() {
    }

    /**
     * split the token into many tokens, much like String.split(), except that the split terms become their
     * own QueryTokens as well.
     * 
     * @param term
     * @return
     */
    public LinkedList<QueryToken> split(String divider) {
        // not implemented by default
        return null;
    }

    /**
     * filter the given set of permutations and return the ones that match this token and any of its children.
     * 
     * @param permutations
     * @return
     * @throws UnknownBeliefLabelException
     */
    public Set<BeliefPermutation> filter(BeliefPermutationIndex index, Set<BeliefPermutation> permutations)
            throws UnknownBeliefLabelException {
        return null; // no implementation by default (we want the NullPointerExceptions if this is ever called)
    }

    public QueryToken copy() {
        QueryToken result = null;
        try {
            result = getClass().newInstance();
        }
        catch (InstantiationException | IllegalAccessException e) {
            // WTF?!
            e.printStackTrace();
        }
        result.children = new HashSet<QueryToken>();
        for (QueryToken child : children) {
            result.children.add(child.copy());
        }
        return result;
    }

    /**
     * return a new node/tree such that query.filter() returns the inverse result set
     */
    public QueryToken inverse() {
        // override where sensible
        return null;
    }

    /**
     * is this token to be used as a conjunction between beliefs or sub-trees? I.e. 'OR' in the query "a OR b"
     * 
     * @return true if the token joins 2 other terms together, false if it cannot
     */
    public boolean isConjunct() {
        return false;
    }

    @Override
    public String toString() {
        // WARNING: this result is used in code and should not be changed.
        // (normally, toString() is used as a debugging-only method)
        return "|" + this.getClass().getSimpleName() + "|";
    }

    public static class And extends QueryToken {

        @Override
        public Set<BeliefPermutation> filter(BeliefPermutationIndex index, Set<BeliefPermutation> permutations)
                throws UnknownBeliefLabelException {

            // AND: progressively filter all children to find the intersection
            Set<BeliefPermutation> result = null;
            for (QueryToken child : children) {
                if (result == null) {
                    result = child.filter(index, permutations);
                }
                else {
                    result = child.filter(index, result);
                }
            }

            return result;
        }

        @Override
        public QueryToken inverse() {
            // not(and(a, b, ...)) == or(not a, not b, ...)
            Or result = new Or();
            for (QueryToken child : children) {
                result.children.add(child.inverse());
            }
            return result;
        }

        @Override
        public boolean isConjunct() {
            return true;
        }
    }

    public static class Or extends QueryToken {

        @Override
        public Set<BeliefPermutation> filter(BeliefPermutationIndex index, Set<BeliefPermutation> permutations)
                throws UnknownBeliefLabelException {

            // OR: filter all children and union the results
            Set<BeliefPermutation> result = new HashSet<BeliefPermutation>();
            for (QueryToken child : children) {
                result.addAll(child.filter(index, permutations));
            }

            return result;
        }

        @Override
        public QueryToken inverse() {
            // not(or(a, b, ...)) == and(not a, not b, ...)
            And result = new And();
            for (QueryToken child : children) {
                result.children.add(child.inverse());
            }
            return result;
        }

        @Override
        public boolean isConjunct() {
            return true;
        }
    }

    public static class Belief extends QueryToken {
        private final String label;
        private final Boolean booleanValue;

        public Belief(String label) {
            if (label.charAt(0) == '-') {
                this.label = label.substring(1);
                this.booleanValue = Boolean.FALSE;
            }
            else {
                this.label = label;
                this.booleanValue = Boolean.TRUE;
            }
        }

        public Belief(String label, Boolean booleanValue) {
            this.label = label;
            this.booleanValue = booleanValue;
        }

        @Override
        public String toString() {
            if (booleanValue) {
                return label + "*";
            }
            else {
                return "-" + label + "*";
            }
        }

        @Override
        public Set<BeliefPermutation> filter(BeliefPermutationIndex index, Set<BeliefPermutation> permutations)
                throws UnknownBeliefLabelException {
            Set<BeliefPermutation> result = new HashSet<BeliefPermutation>(permutations);
            result.retainAll(index.getBeliefPermutations(label, booleanValue));
            return result;
        }

        @Override
        public QueryToken inverse() {
            // a == false -> not(a) == true
            // a == true -> not(a) == false
            Belief result = new Belief(label, !booleanValue);
            return result;
        }
    }

    public void addChild(QueryToken token) {
        children.add(token);
    }

    public static class StartGroup extends QueryToken {
    }

    public static class EndGroup extends QueryToken {
    }

    public static class UnknownToken extends QueryToken {
        private String term = null;

        public UnknownToken(String term) {
            this.term = term;
        }

        @Override
        public LinkedList<QueryToken> split(String divider) {
            LinkedList<QueryToken> result = new LinkedList<QueryToken>();
            String subject = new String(this.term);

            int splitLocation;
            while (true) {
                splitLocation = subject.indexOf(divider);
                if (splitLocation == -1) {
                    break;
                }
                else if (splitLocation == 0) {
                    result.add(new UnknownToken(divider));
                    subject = subject.substring(divider.length());
                }
                else {
                    result.add(new UnknownToken(subject.substring(0, splitLocation)));
                    subject = subject.substring(splitLocation);
                }
            }

            // add whatever was after the last divider
            if (subject.length() > 0) {
                result.add(new UnknownToken(subject));
            }
            return result;
        }

        @Override
        public String toString() {
            // WARNING: this result is used in code and should not be changed.
            // (normally, toString() is used as a debugging-only method)
            return new String(term);
        }

    }
}
