package queries;

import exceptions.QueryAmbiguityException;
import exceptions.QueryConjunctPairingException;
import exceptions.QueryFormattingException;
import queries.QueryToken.UnknownToken;

/**
 * Linked list implementation specifically for use during query parsing
 * 
 * @author luke
 */
public class TokenList {
    // package-private for easier use
    TokenListNode first = null;
    TokenListNode last = null;

    public void addAll(String[] terms) {
        for (String term : terms) {
            add(new UnknownToken(term));
        }
    }

    /**
     * add a single QueryToken to the list, wrapping it in a TokenListNode.
     * 
     * @param node
     */
    public void add(QueryToken token) {
        TokenListNode node = new TokenListNode(token);
        add(node);
    }

    /**
     * add one or many nodes to this list
     * 
     * @param node
     */
    public void add(TokenListNode node) {
        if (first == null) {
            // we're adding the first node
            first = node;
            node.previous = null;
        }
        else {
            // we're appending to an existing list of nodes
            node.previous = last;
            last.next = node;
        }
        // if we're adding multiple nodes, find the last one and link to it. 
        while (node.next != null) {
            node = node.next;
        }
        last = node;
    }

    /**
     * insert a list of new nodes BEFORE the given QueryToken.
     * 
     * @param insertAtToken
     * @param insertingNode
     * @throws TokenNotFoundException
     */
    public void insertNodes(QueryToken insertAtToken, TokenListNode insertingNode)
            throws TokenNotFoundException {

        TokenListNode insertLocation = seek(insertAtToken);

        insertingNode.previous = insertLocation.previous;
        if (insertLocation.previous == null) {
            // we're inserting at the first node
            first = insertingNode;
        }
        else {
            // we're inserting somewhere in the middle
            insertLocation.previous.next = insertingNode;
        }

        // seek to the end of the inserting nodes
        TokenListNode lastInsertingNode = insertingNode;
        while (lastInsertingNode.next != null) {
            lastInsertingNode = lastInsertingNode.next;
        }

        lastInsertingNode.next = insertLocation;
        insertLocation.previous = lastInsertingNode;
    }

    /**
     * delete the given QueryToken from the list
     * 
     * @param token
     * @throws TokenNotFoundException
     */
    public void delete(QueryToken token) throws TokenNotFoundException {
        TokenListNode deleteNode = seek(token);
        if (deleteNode.previous == null) {
            // we're deleting the first node
            first = deleteNode.next;
            first.previous = null;
        }
        else {
            deleteNode.previous.next = deleteNode.next; // will be null if we're
                                                        // deleting the last
                                                        // node
            if (deleteNode.next == null) {
                // we're deleting the last node
                last = deleteNode.previous;
            }
            else {
                deleteNode.next.previous = deleteNode.previous;
            }
        }
    }

    private TokenListNode seek(QueryToken token) throws TokenNotFoundException {
        TokenListNode insertLocation = first;
        while (insertLocation.token != token) {
            insertLocation = insertLocation.next;
            if (insertLocation == null) {
                throw new TokenNotFoundException();
            }
        }
        return insertLocation;
    }

    /**
     * Create a tree of the nodes in this list (including recursive calls in the case that a node is a
     * TokenListNodeGroup)
     * 
     * @return
     * @throws QueryFormattingException
     */
    public QueryToken asTree() throws QueryFormattingException {
        // a list should be in the form a-AND-b-AND-c... or a-OR-b-OR-c...
        // which is a tree rooted at AND or OR, with a,b,c... as children,
        // where a,b,c... are either plain beliefs (QueryToken.Belief),
        // or subtrees (rooted at either AND or OR)
        // this implies that there is an odd/even pattern, 
        // where every other token is a,b,c... or AND/OR.

        TokenListNode firstNode = first;
        TokenListNode secondNode = getSecondNode(firstNode);
        if (secondNode == null) {
            // there is only 1 term, so return it as the root node.
            return first.asTree(); // 2-step recursion
        }

        // there are multiple terms, so make a root node that joins them.
        QueryToken root = secondNode.token;
        root.addChild(firstNode.asTree());

        while (true) {
            firstNode = secondNode.next;
            secondNode = getSecondNode(firstNode);

            root.addChild(firstNode.asTree()); // 2-step recursion
            if (secondNode == null) {
                // end of the list. 
                break;
            }

            // make sure conjuncts are all the same type (a AND b OR c is ambiguous!)
            if (secondNode.token.getClass() != root.getClass()) {
                throw new QueryAmbiguityException();
            }
            // we can throw away each secondNode because it matches the root and doesn't add any new info.

            // done. go to the next position.
        }

        return root;
    }

    private TokenListNode getSecondNode(TokenListNode firstNode) throws QueryConjunctPairingException {
        if (firstNode == null) {
            // got an even number of terms (not possible!)
            throw new QueryConjunctPairingException();
        }
        TokenListNode secondNode = firstNode.next;
        if (secondNode == null) {
            return null;
        }
        if (secondNode != null && !secondNode.isConjunct()) {
            // the second node is not AND or OR, so it cannot be used as a conjunct.
            throw new QueryConjunctPairingException();
        }
        return secondNode;
    }

    @Override
    public String toString() {
        String result = "[";
        TokenListNode current = first;
        while (current != null) {
            result += current.toString() + ",";
            current = current.next;
        }
        if (result.length() > 1) {
            result = result.substring(0, result.length() - 1);
        }
        result += "]";
        return result;
    }
}
