package revisionOperators;

import java.lang.reflect.Constructor;
import java.util.Set;


import queries.Query;
import datastructures.BeliefPermutation;
import datastructures.BeliefPermutationIndex;
import datastructures.CumulativeMeasure;
import datastructures.Value;
import exceptions.IllegalRevisionException;
import exceptions.QueryFormattingException;
import exceptions.UnknownBeliefLabelException;

public abstract class RevisionConstraint {
    protected Query query;
    protected Value value = null; // will be left null until revise() if using a valueQuery
    protected Query valueQuery = null; // leave null unless the query is in the format A <= B

    /**
     * static constructor method - use regex to decide which kind of comparison/revision is being done.
     * (options: <=, =, >=)
     * 
     * @param queryString
     * @return
     * @throws QueryFormattingException
     */
    public static RevisionConstraint fromQueryString(String queryString) throws QueryFormattingException {
        String splitString = null;
        Class<? extends RevisionConstraint> resultClass = null;
        if (queryString.matches(".*<=.*")) {
            splitString = "<=";
            resultClass = LessThanOrEqualTo.class;
        } else if (queryString.matches(".*>=.*")) {
            splitString = ">=";
            resultClass = GreaterThanOrEqualTo.class;
        } else if (queryString.matches(".*=.*")) {
            splitString = "=";
            resultClass = EqualTo.class;
        } else {
            Query query = new Query(queryString); // a query that returns the set to be "believed in" (which may be in the form "-a")
            Value value = new Value(1, 1.0); // a value representing "not believed in"
            RevisionConstraint result = new LessThanOrEqualTo(query.inverse(), value); // i.e. "-a is not believed in"
            return result;
        }

        String criteriaString = queryString.split(splitString)[0];
        String valueString = queryString.split(splitString)[1];

        Constructor<? extends RevisionConstraint> constructor = null;
        try {
            constructor = resultClass.getConstructor(Query.class, Value.class);
        } catch (NoSuchMethodException | SecurityException e) {
            // WTF?!
            e.printStackTrace();
        }
        Query queryArg = new Query(criteriaString);
        try {
            Value valueArg = new Value(valueString);
            return constructor.newInstance(queryArg, valueArg);
        } catch (Exception e) {
            // valueString is probably a sub query, not a Value.
            try {
                constructor = resultClass.getConstructor(Query.class, Query.class);
                Query valueQuery = new Query(valueString);
                return constructor.newInstance(queryArg, valueQuery);
            } catch (Exception e2) {
                // WTF?!
                e.printStackTrace();
            }
        }
        return null;
    }

    public RevisionConstraint(String queryString, String valueString) throws QueryFormattingException {
        this.query = new Query(queryString);
        this.value = new Value(valueString);
    }

    public RevisionConstraint(Query query, Value value) {
        this.query = query;
        this.value = value;
    }

    public RevisionConstraint(Query constrainingQuery, Query valueQuery) {
        this.query = constrainingQuery;
        this.valueQuery = valueQuery;
    }

    public Set<BeliefPermutation> filter(Set<BeliefPermutation> permutations, BeliefPermutationIndex index)
            throws UnknownBeliefLabelException {
        return query.filter(permutations, index);
    }

    public abstract Value revise(Value value);

    /**
     * Apply this revision constraint to a Cumulative Measure by scaling the BeliefPermutations that match the
     * query.
     * 
     * This is the main algorithm in the Cumulative Measure tool kit.
     * 
     * @param cumulativeMeasure
     * @throws UnknownBeliefLabelException
     * @throws IllegalRevisionException
     */
    public void revise(CumulativeMeasure cumulativeMeasure) throws UnknownBeliefLabelException,
            IllegalRevisionException {

        setConstraintValueIfNecessary(cumulativeMeasure);

        Set<BeliefPermutation> allPermutations = cumulativeMeasure.getBeliefPermutationsCopy();

        // identify the set of permutations that will be changed (the "revision set")
        Set<BeliefPermutation> revisionSet = query.filter(allPermutations, cumulativeMeasure.getIndex());

        // determine the final constrained value that the revision set will have
        Value revisionSum = Value.sum(revisionSet);
        Value targetValue = revise(revisionSum);

        // terminate early if there is nothing to do
        if (revisionSum.equals(targetValue)) {
            return;
        }

        // Multiplicative accessibility doesn't work for infinity.
        if (revisionSum.equals(Value.min())) {
            throw new IllegalRevisionException();
        }

        // calculate a multiplier that must be used to achieve the target value (can be infinity)
        Value multiplier = targetValue.dividedBy(revisionSum);

        // scale all the values in the revision set to achieve the target value
        for (BeliefPermutation permutation : revisionSet) {
            Value newValue = permutation.valueCopy().multipliedBy(multiplier);
            permutation.setValue(newValue);
        }

        // scale local values of the stable set to accommodate rank 0 (if necessary)
        if (!Value.sum(allPermutations).equals(Value.max())) {
            // rank 0 is not scaled properly to add up to (0,1), so scale it!

            // identify the set of permutations that will not be changed (the "stable set")
            Set<BeliefPermutation> stableSet = query.inverse().filter(allPermutations,
                    cumulativeMeasure.getIndex());

            // calculate a multiplier that must be used to achieve the target value
            revisionSum = Value.sum(revisionSet);
            if (revisionSum.lessThan(new Value(0, 0.0))) {
                // the revision set does not occupy rank 0, so the stable set will have to sum up to (0,1) - the full rank-1 amount.
                targetValue = Value.max();
            } else {
                // rank 0 will be shared by BeliefPermutations of both the stable set and the revision set
                targetValue = new Value(0, 1.0 - revisionSum.local());
            }

            Value stableSum = Value.sum(stableSet);
            if (stableSum.equals(Value.min())) {
                // Multiplicative accessibility doesn't work for infinity.
                throw new IllegalRevisionException();
            }
            multiplier = targetValue.dividedBy(stableSum);

            // scale all the values in the stable set to achieve the target value
            for (BeliefPermutation permutation : stableSet) {
                Value newValue = permutation.valueCopy().multipliedBy(multiplier);
                permutation.setValue(newValue);
            }
        }

        // sanity check
        if (!Value.sum(allPermutations).equals(Value.max())) {
            throw new IllegalRevisionException();
        }

        // everything worked. Apply the changes.
        cumulativeMeasure.setBeliefPermutations(allPermutations);
    }

    private void setConstraintValueIfNecessary(CumulativeMeasure cumulativeMeasure)
            throws UnknownBeliefLabelException {
        if (value == null) {
            value = Value.sum(cumulativeMeasure.query(valueQuery));
        }
    }
}
