package test;

import static org.junit.Assert.assertTrue;

import org.junit.Test;

import exceptions.IllegalRevisionException;
import exceptions.QueryFormattingException;
import exceptions.UnknownBeliefLabelException;

/**
 * Test scenarios inspired by: A. Darwiche and J. Pearl. On the logic of iterated belief revision. Artificial
 * intelligence, 89(1):1–29, 1997
 */
public class Literature_DarwicheTests extends TestBase {

    /**
     * "Red Bird" example from Page 9
     * 
     * @throws IllegalRevisionException
     * @throws UnknownBeliefLabelException
     * @throws QueryFormattingException
     */
    @Test
    public void redBirdTest() throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        cumulative.add("Bird");
        cumulative.add("Red");

        cumulative.revise("Bird");
        assertTrue(value("Red AND Bird", 0, 0.5));
        assertTrue(value("-Red AND Bird", 0, 0.5));
        assertTrue(value("Red AND -Bird", 1, 0.5));
        assertTrue(value("-Red AND -Bird", 1, 0.5));

        cumulative.revise("Red");
        assertTrue(value("Red AND Bird", 0, 1.0));
        assertTrue(value("-Red AND Bird", 1, 1.0));
        assertTrue(value("Red AND -Bird", 1, 1.0));
        assertTrue(value("-Red AND -Bird", 2, 1.0));

        cumulative.revise("-Bird");
        assertTrue(value("Red AND -Bird", 0, 1.0)); // AGM postulate CB would have us believe -Red
        assertTrue(value("-Red AND -Bird", 1, 1.0));
        assertTrue(value("Red AND Bird", 1, 1.0));
        assertTrue(value("-Red AND Bird", 2, 1.0));
    }

    /**
     * "Johnny and Mary" example on (p.15)
     * 
     * @throws IllegalRevisionException
     * @throws UnknownBeliefLabelException
     * @throws QueryFormattingException
     */
    @Test
    public void johnAndMaryTest() throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        // suppose that we face a murder trial with two main suspects, John and Mary. 
        cumulative.add("John");
        cumulative.add("Mary");

        // Initially, it appears that the murder was committed by one person, hence, we believe that
        cumulative.revise("(John AND -Mary) OR (-John AND Mary)");
        assertTrue(value("John AND -Mary", 0, 0.5));
        assertTrue(value("-John AND Mary", 0, 0.5));
        assertTrue(value("John AND Mary", 1, 0.5));
        assertTrue(value("-John AND -Mary", 1, 0.5));
        // Given the AGM postulates, we also believe in the two conditionals:
        // -Mary|John and ~John|Mary

        // As the trial unfolds, however, we receive a very reliable testimony incriminating John...
        cumulative.revise("John");
        assertTrue(value("John AND -Mary", 0, 1.0));
        assertTrue(value("-John AND Mary", 1, 1.0));
        assertTrue(value("John AND Mary", 1, 1.0));
        assertTrue(value("-John AND -Mary", 2, 1.0));

        // ...followed by another reliable testimony incriminating Mary.
        cumulative.revise("Mary");

        // At this point, it is only reasonable to believe that both suspects 
        // took part in the murder, thus dismissing the one-person theory
        // together with the two conditional beliefs.
        assertTrue(value("-John AND Mary", 0, 0.5)); // see below (A)
        assertTrue(value("John AND Mary", 0, 0.5)); // see below (B)
        assertTrue(value("John AND -Mary", 1, 1.0));
        assertTrue(value("-John AND -Mary", 3, 1.0));

        /*
         * the commented lines show that Cumulative Measures will consider our "single murderer" theory (A) as
         * reliable as the more recent information that there is evidence against both John and Mary (B)
         */
    }
}
