package test;

import static org.junit.Assert.assertTrue;

import org.junit.Test;

import exceptions.IllegalRevisionException;
import exceptions.QueryFormattingException;
import exceptions.UnknownBeliefLabelException;

/**
 * Test scenarios inspired by: S.M. Glaister. Symmetry and belief revision. Erkenntnis, 49(1):21–56, 1998.
 * (specifically, examples 1.9 and 1.10 on p.31)
 * 
 */
public class Literature_GlaisterTests extends TestBase {

    @Test
    public void test1_9() throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        cumulative.add("A");
        cumulative.add("B");

        //I believe that both that my friend Andrew is married (A) and that he is
        //living in California (B). Moreover these are independent commitments for
        //me, indeed, my epistemic state is representable as
        //<1 = |A&B| < |A&-B| < |-A&B| < |-A&-B|

        // NOTE: to accomplish the starting state, I'm using a bit of a syntax trick:
        // first, assert belief in the LEAST believed thing, so that all other values are pushed down to rank 1.
        // then, proceed in a round-robin fashion, asserting the other beliefs "on top" of the less believed thing(s).
        cumulative.revise("-A AND -B");
        cumulative.revise("-A AND B");
        cumulative.revise("A AND -B");
        cumulative.revise("A AND B");
        assertTrue(value("A AND B", 0, 1.0));
        assertTrue(value("A AND -B", 1, 1.0));
        assertTrue(value("-A AND B", 2, 1.0));
        assertTrue(value("-A AND -B", 3, 1.0));

        // I learn that Andrew isn't married. M-revising we get:
        // <2 = revise(<1, |~A|) = |~A&B| < |A&B| < |A&~B| < |~A&B|
        cumulative.revise("-A");
        assertTrue(value("-A AND B", 0, 1.0));
        assertTrue(value("-A AND -B", 1, 1.0));
        assertTrue(value("A AND B", 1, 1.0));
        assertTrue(value("A AND -B", 2, 1.0));

        //Now I learn that Andrew isn't living in California either. M-revising once
        //more yields
        //<3 = revise(<2, |~B|) = |A&~B| < |~A&~B| < |A&B| < |~A&B|
        //NOTE: typo was fixed in 2nd term, from |~A&B| to |~A&~B|.
        cumulative.revise("-B");
        assertTrue(value("-A AND -B", 0, 1.0));
        assertTrue(value("A AND -B", 1, 1.0));
        assertTrue(value("-A AND B", 1, 1.0));
        assertTrue(value("A AND B", 2, 1.0));
    }

    @Test
    public void test1_10a() throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        cumulative.add("Mu");
        cumulative.add("ElvisIsAlive");

        // I have just accepted Mu. Now I'm confronted with overwhelming evidence
        //that if Mu then Elvis is alive. Hence I decide to accept...
        cumulative.revise("Mu");
        cumulative.revise("-Mu OR (Mu AND ElvisIsAlive)"); // if Mu then ElvisIsAlive
        assertTrue(value("Mu AND ElvisIsAlive", 0, 1.0)); // forced to conclude that Elvis is alive
        assertTrue(value("-Mu AND ElvisIsAlive", 1, 1.0));
        assertTrue(value("-Mu AND -ElvisIsAlive", 1, 1.0));
        assertTrue(value("Mu AND -ElvisIsAlive", 1, 1.0));
    }
}
