package test;

import static org.junit.Assert.*;

import org.junit.Test;

import datastructures.Value;

import exceptions.IllegalRevisionException;
import exceptions.QueryFormattingException;
import exceptions.UnknownBeliefLabelException;

public class ProjectCompletionTests extends TestBase {

    @Test
    public void progressiveRevision() throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        cumulative.add("a");
        cumulative.add("b");

        // All tests are transcribed directly from an email received from Dr. Emil Weydert (Jan. 4)
        // The comments are exactly as written, with name replacements and whitespace formatting only.

        // expected results are given for belief tuples:
        // ( a b)   ( a -b)
        // (-a b)   (-a -b)
        // using Cumulative Measure notation (GlobalRank, LocalBeliefValue)

        // assert starting point
        // (0, 1/4) (0, 1/4)
        // (0, 1/4) (0, 1/4)
        assertTrue(value("a AND b", 0, 0.25));
        assertTrue(value("-a AND b", 0, 0.25));
        assertTrue(value("a AND -b", 0, 0.25));
        assertTrue(value("-a AND -b", 0, 0.25));

        // * [v(a) >= (0, 1/3)] =
        // (0, 1/4) (0, 1/4)
        // (0, 1/4) (0, 1/4)
        cumulative.revise("a >= (0, 0.3333)");
        assertTrue(value("a AND b", 0, 0.25));
        assertTrue(value("-a AND b", 0, 0.25));
        assertTrue(value("a AND -b", 0, 0.25));
        assertTrue(value("-a AND -b", 0, 0.25));

        // * [v(a AND b) >= (0, 1/2)] =
        // (0, 1/2) (0, 1/6)
        // (0, 1/6) (0, 1/6)
        cumulative.revise("a AND b >= (0, 0.5)");
        assertTrue(value("a AND b", 0, 0.5));
        assertTrue(value("-a AND b", 0, 0.1667));
        assertTrue(value("a AND -b", 0, 0.1667));
        assertTrue(value("-a AND -b", 0, 0.1667));

        // * [v(-b) <= (1, 4)] =
        // (0, 3/4) (1, 2)
        // (0, 1/4) (1, 2)
        cumulative.revise("-b <= (1, 4.0)");
        assertTrue(value("a AND b", 0, 0.75));
        assertTrue(value("-a AND b", 0, 0.25));
        assertTrue(value("a AND -b", 1, 2));
        assertTrue(value("-a AND -b", 1, 2));

        // * [v(-a) <= (1, 4)] =
        // (0, 1) (1, 8/3)
        // (1, 4) (2, 32)
        cumulative.revise("-a <= (1, 4.0)");
        assertTrue(value("a AND b", 0, 1.0));
        assertTrue(value("-a AND b", 1, 4.0));
        assertTrue(value("a AND -b", 1, 2.6666666));
        assertTrue(value("-a AND -b", 2, 32));

        // * [v(-b) >= (0, 1/6)] =
        // (0, 5/6) (0, 1/6)
        // (1, 10/3) (1, 2)
        cumulative.revise("-b >= (0, 0.166666)");
        assertTrue(value("a AND b", 0, 0.833333));
        assertTrue(value("-a AND b", 1, 3.333333));
        assertTrue(value("a AND -b", 0, 0.166666));
        assertTrue(value("-a AND -b", 1, 2));

        // * [v(a) <= (1, 1)] =
        // (1, 5/6) (1, 1/6)
        // (0, 5/8) (0, 3/8)
        cumulative.revise("a <= (1,1)");
        assertTrue(value("a AND b", 1, 0.833333));
        assertTrue(value("-a AND b", 0, 0.625));
        assertTrue(value("a AND -b", 1, 0.166666));
        assertTrue(value("-a AND -b", 0, 0.375));

        // * [v(a AND -b) >= (1, 1)] =
        // (1, 5/6) (1, 1)
        // (0, 5/8) (0, 3/8)
        cumulative.revise("a AND -b >= (1,1)");
        assertTrue(value("a AND b", 1, 0.833333));
        assertTrue(value("-a AND b", 0, 0.625));
        assertTrue(value("a AND -b", 1, 1));
        assertTrue(value("-a AND -b", 0, 0.375));

        // * [v(-b) <= (INF, 0)] =
        // (1, 4/3) (INF, 0)
        // (0, 1) (INF, 0)
        cumulative.revise("-b <= (INF,0)");
        assertTrue(value("a AND b", 1, 1.333333));
        assertTrue(value("-a AND b", 0, 1.0));
        assertTrue(value("a AND -b", Value.GLOBAL_INFINITY, 0.0));
        assertTrue(value("-a AND -b", Value.GLOBAL_INFINITY, 0.0));

        // IMPOSSIBLE REVISION! 
        // Should throw an exception for normal revise(),
        // or do nothing for reviseSafe().
        // * [v(b) <= (INF, 0)] =
        // (1, 4/3) (INF, 0)
        // (0, 1) (INF, 0)
        try {
            cumulative.revise("b <= (INF,0)");
            fail("Expected IllegalRevisionException");
        }
        catch (IllegalRevisionException e) {
            // pass
        }
        cumulative.reviseSafe("b <= (INF,0)");
        assertTrue(value("a AND b", 1, 1.333333));
        assertTrue(value("-a AND b", 0, 1.0));
        assertTrue(value("a AND -b", Value.GLOBAL_INFINITY, 0.0));
        assertTrue(value("-a AND -b", Value.GLOBAL_INFINITY, 0.0));

        // * [v(a AND b) >= (0, 2/3)] =
        // (0, 2/3) (INF, 0)
        // (0, 1/3) (INF, 0)
        cumulative.revise("a AND b >= (0, 0.666666666)");
        assertTrue(value("a AND b", 0, 0.666666));
        assertTrue(value("-a AND b", 0, 0.3333333));
        assertTrue(value("a AND -b", Value.GLOBAL_INFINITY, 0.0));
        assertTrue(value("-a AND -b", Value.GLOBAL_INFINITY, 0.0));
    }

    @Test
    public void plainBelief() throws QueryFormattingException, UnknownBeliefLabelException,
            IllegalRevisionException {
        cumulative.add("a");
        cumulative.add("b");

        // All tests are written directly from an email received from Dr. Emil Weydert
        // The comments are exactly as written, with name replacements and whitespace formatting only.

        // expected results are given for belief tuples:
        // ( a b)   ( a -b)
        // (-a b)   (-a -b)
        // using Cumulative Measure notation (GlobalRank, LocalBeliefValue)

        // assert starting point
        // (0, 1/4) (0, 1/4)
        // (0, 1/4) (0, 1/4)
        assertTrue(value("a AND b", 0, 0.25));
        assertTrue(value("-a AND b", 0, 0.25));
        assertTrue(value("a AND -b", 0, 0.25));
        assertTrue(value("-a AND -b", 0, 0.25));

        // revise by plain belief in a
        // (0, 1/2) (0, 1/2)
        // (1, 1/2) (1, 1/2)
        cumulative.revise("a");
        assertTrue(value("a AND b", 0, 0.5));
        assertTrue(value("-a AND b", 1, 0.5));
        assertTrue(value("a AND -b", 0, 0.5));
        assertTrue(value("-a AND -b", 1, 0.5));

    }
}
