package test;

import static org.junit.Assert.*;

import datastructures.BeliefPermutation;
import datastructures.CumulativeMeasure;
import datastructures.Value;

import org.junit.Before;
import org.junit.Test;


import exceptions.IllegalRevisionException;
import exceptions.QueryFormattingException;
import exceptions.UnknownBeliefLabelException;

public class QueryTests extends TestBase {

    private BeliefPermutation ab;
    private BeliefPermutation Ab;
    private BeliefPermutation aB;
    private BeliefPermutation AB;

    private BeliefPermutation abc;
    private BeliefPermutation abC;
    private BeliefPermutation aBc;
    private BeliefPermutation aBC;
    private BeliefPermutation Abc;
    private BeliefPermutation AbC;
    private BeliefPermutation ABc;
    private BeliefPermutation ABC;

    @Override
    @Before
    public void setUp() throws Exception {
        cumulative = new CumulativeMeasure();

        Value v0_025 = v(0, 0.25);
        ab = new BeliefPermutation();
        ab.addBelief("a", false);
        ab.addBelief("b", false);
        ab.setValue(v0_025);

        aB = new BeliefPermutation();
        aB.addBelief("a", false);
        aB.addBelief("b", true);
        aB.setValue(v0_025);

        Ab = new BeliefPermutation();
        Ab.addBelief("a", true);
        Ab.addBelief("b", false);
        Ab.setValue(v0_025);

        AB = new BeliefPermutation();
        AB.addBelief("a", true);
        AB.addBelief("b", true);
        AB.setValue(v0_025);

        Value v0_0125 = v(0, 0.125);
        abc = new BeliefPermutation();
        abc.addBelief("a", false);
        abc.addBelief("b", false);
        abc.addBelief("c", false);
        abc.setValue(v0_0125);

        abC = new BeliefPermutation();
        abC.addBelief("a", false);
        abC.addBelief("b", false);
        abC.addBelief("c", true);
        abC.setValue(v0_0125);

        aBc = new BeliefPermutation();
        aBc.addBelief("a", false);
        aBc.addBelief("b", true);
        aBc.addBelief("c", false);
        aBc.setValue(v0_0125);

        aBC = new BeliefPermutation();
        aBC.addBelief("a", false);
        aBC.addBelief("b", true);
        aBC.addBelief("c", true);
        aBC.setValue(v0_0125);

        Abc = new BeliefPermutation();
        Abc.addBelief("a", true);
        Abc.addBelief("b", false);
        Abc.addBelief("c", false);
        Abc.setValue(v0_0125);

        AbC = new BeliefPermutation();
        AbC.addBelief("a", true);
        AbC.addBelief("b", false);
        AbC.addBelief("c", true);
        AbC.setValue(v0_0125);

        ABc = new BeliefPermutation();
        ABc.addBelief("a", true);
        ABc.addBelief("b", true);
        ABc.addBelief("c", false);
        ABc.setValue(v0_0125);

        ABC = new BeliefPermutation();
        ABC.addBelief("a", true);
        ABC.addBelief("b", true);
        ABC.addBelief("c", true);
        ABC.setValue(v0_0125);
    }

    @Test
    public void simple() throws QueryFormattingException, UnknownBeliefLabelException {
        cumulative.add("a");
        cumulative.add("b");

        assertTrue(result("a AND b", new BeliefPermutation[] { AB }));
        assertTrue(result("a AND -b", new BeliefPermutation[] { Ab }));
        assertTrue(result("-a AND b", new BeliefPermutation[] { aB }));
        assertTrue(result("-a AND -b", new BeliefPermutation[] { ab }));
    }

    @Test
    public void complex() throws QueryFormattingException, UnknownBeliefLabelException {
        cumulative.add("a");
        cumulative.add("b");
        cumulative.add("c");

        // abc, abC, aBc, aBC, Abc, AbC, ABc, ABC
        assertTrue(result("a OR b", new BeliefPermutation[] { Abc, aBc, ABc, AbC, aBC, ABC }));
        assertTrue(result("a OR c", new BeliefPermutation[] { Abc, AbC, abC, ABc, ABC, aBC }));
        assertTrue(result("a OR -c", new BeliefPermutation[] { Abc, abc, AbC, ABc, aBc, ABC }));
        assertTrue(result("-a AND (b OR c)", new BeliefPermutation[] { abC, aBc, aBC }));
        assertTrue(result("(-a OR -b) AND c", new BeliefPermutation[] { abC, aBC, AbC }));
        assertTrue(result("(-a AND -b) OR (c AND -a) OR (b AND c)", new BeliefPermutation[] { abc, abC,/* abC, */
        aBC, /* aBC, */ABC }));
    }

    @Test
    public void formattingExceptions() throws UnknownBeliefLabelException, IllegalRevisionException,
            QueryFormattingException {
        cumulative.add("a");
        cumulative.add("b");
        cumulative.add("c");

        cumulative.query("a  AND  (\t\tb\tOR    c)"); // space collapsing & tabs - shouldn't throw
        assertTrue(queryThrowsFormattingException("a <= 0,1)")); // missing )
        assertTrue(queryThrowsFormattingException("a <= (a,1)")); // char
    }

    private boolean queryThrowsFormattingException(String queryString) throws UnknownBeliefLabelException {
        try {
            cumulative.query(queryString);
            return false;
        } catch (QueryFormattingException e) {
            // pass
            return true;
        }
    }

    @Test
    public void conjunctExceptions() throws QueryFormattingException, UnknownBeliefLabelException {
        cumulative.add("a");
        cumulative.add("b");
        cumulative.add("c");

        assertTrue(badConjunct("(-a AND -b) OR (c AND -a) OR (b and c)")); // lower case 'and'
        assertTrue(badConjunct("(-a AND -b) OR (c AND -a) or (b AND c)")); // lower case 'or'
        assertTrue(badConjunct("(-a AND -b) OR OR (c AND -a) OR (b AND c)")); // OR OR 
        assertTrue(badConjunct("OR (c AND -a) OR (b AND c)")); //starts with OR
        assertTrue(badConjunct("AND (b AND c)")); // starts with AND
        assertTrue(badConjunct("(-a AND -b) OR (c AND -a) OR (b AND c) OR")); // ends with OR
    }

    @Test
    public void ambiguityExceptions() throws QueryFormattingException, UnknownBeliefLabelException {
        cumulative.add("a");
        cumulative.add("b");
        cumulative.add("c");

        assertTrue(ambiguous("a AND b OR c")); // the basic form
        assertTrue(ambiguous("a OR b AND c")); // the basic form
        assertTrue(ambiguous("(-a AND -b) OR (c AND -a) AND (b AND c)"));
        assertTrue(ambiguous("(-a AND -b OR c) AND (c AND -a) AND (b AND c)"));
    }
}
